<?php
/****************************************************************************
 *                                                                          *
 *                         P A B [PHPAntiBot]                               *
 *                                                                          *
 * Copyright (C) 2006-2007 Jonathan BURON <webmaster@natsimhan.com>         *
 *                                                                          *
 * This file is part of PAB                                                 *
 *                                                                          *
 * PHPAntiBot is free software; you can redistribute it and/or modify       *
 * it under the terms of the GNU General Public License as published by     *
 * the Free Software Foundation; either version 2 of the License.           *
 *                                                                          *
 * PHPAntiBot is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             *
 * GNU General Public License for more details.                             *
 *                                                                          *
 * You should have received a copy of the GNU General Public License        *
 * along with PHPAntiBot; if not, write to the Free Software                *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA *
 ****************************************************************************/

# Cette classe est capable de généré une image contenant un string déformé 
# et illisible pour un OCR dans la mesure du possible et permet donc dans l'utilisation
# d'interdire à un robot du net de valider un formulaire automatiquement
class PAB {
  private static $_Instance = null;
  
  const VERSION = "0.1";     # Version de l'application PAB

  private $_TabPK = array();  # Tableau contenant les valeurs des clées publiques/privées
  private $_Bordure=1;        # Taille de la bordure générée, 0=>pas de bordure
  private $_MargeLettre=1;
  private $_AngleMax=30;
  private $_CopyResampled=true;
  private $_FontType=5;
  private $_margeImage=10;
  private $_hauteurImage=50;
  private $_SenssibleCasse=false;
  
  # Constructeur de la classe
  private function __construct() {
    # Initialisation du tableau des clées disponibles
    $this->keyFileToTab();
	}
  # Récupère une instance unique de la classe
  public static function getInstance($debug=false){
    if($debug){
      error_reporting(E_ALL ^ E_STRICT);
      ini_set('display_errors',true);
    }
    if(self::$_Instance === null) {
      self::$_Instance = new PAB();
    }
    return self::$_Instance;
  }
  # Cette function génère l'image en question et retourne un pointeur image PHP
  public function getImage() {
    # On récupère la clée privée d'après la clée publique
    // --- On récupère la clée publique définit dans le formulaire
    if(!isset($_GET['PublicKey'])) return imagecreatefrompng('images/no_cle_publique.png');
    $public_key = $_GET['PublicKey'];
    if(!settype($public_key,"string")) return imagecreatefrompng('images/no_cle_publique.png');
    $private_key = $this->getPrivateKey($public_key);
    # On sépare tous les caractères de la chaine de caractère afin de les traiter individuellement
    $tb_char = preg_split('//', $private_key, -1, PREG_SPLIT_NO_EMPTY);
    # On créé l'image finale qui va contenir les caractères tournés
    $im_des =  imagecreatetruecolor($this->_margeImage,$this->_hauteurImage);
    $white = imagecolorallocate($im_des, 255, 255, 255);
    imagefill($im_des,0,0,$white);
    # On définit la taille del a police utilisée
    $textsizemin = 18;
    $textsizemax = 26;
    # Nom de la police à utiliser
    $fontfiles = array('alanden','luggerbu','wavy');
    $fontfile = 'fonts/'.$fontfiles[mt_rand(0,count($fontfiles)-1)].'.ttf';
    # On boucle sur la totalité des caractères
    foreach($tb_char as $char){
      # Calcul aléatoire de la rotation du caractère
      $angle = mt_rand(0, 2*$this->getAngleMax())-$this->getAngleMax();
      # calcul aléatoire de la taille du caractère
      $textsize = mt_rand($textsizemin, $textsizemax);
      # on récupère les dimensions de l'image du caractère
      $rec = $this->myImagettfbbox($textsize, $angle, $fontfile, $char);
      $w = $rec['width'];
      $h = $rec['height'];
      # on récupère les dimensions de l'image finale
      $w_des = imagesx($im_des);
      $h_des = imagesy($im_des);
      # on calcul les nouvelles dimensions
      $w_new = $w_des + $w + 2;
      # on cré l'image temporaire dans laquelle on colle l'image finale puis le caractère
      $im_tmp = imagecreatetruecolor($w_new,$this->_hauteurImage);
      # on définit la couleur de fond et on colorie le fond avec
      $white = imagecolorallocate($im_tmp, 255, 255, 255);
      imagefill($im_tmp,0,0,$white);
      # on copie l'ancienne image dans la nouvelle
      imagecopy($im_tmp,$im_des,0,0,0,0,$w_des,$this->_hauteurImage);
      # on définit la couleur du caractère
      $color = imagecolorallocate($im_tmp, 0x00,0x00,0x00);
      # on dessine le caractère dans la nouvelle image
      imagettftext($im_tmp, $textsize, $angle, $w_des, $this->_margeImage+(($this->_hauteurImage-$this->_margeImage-$this->_margeImage)/1.5), $color, $fontfile, $char);
      # puis on enregiste l'image temp dans celle de destination
      imagedestroy($im_des);
      $im_des = $im_tmp;

    }
    
    $color_blanc = imagecolorallocate($im_des, 0xff, 0xff, 0xff);
    $color_noir = imagecolorallocate($im_des, 0,0,0);
    $color_rouge = imagecolorallocate($im_des, 0xff,0,0);
    $color_vert = imagecolorallocate($im_des, 0, 0xff,0);
    $color_bleu = imagecolorallocate($im_des, 0, 0, 0xff);

    $this->traceGrille($im_des);

    # on récupère les dimensions de l'image finale
    $w_des = imagesx($im_des);
    $h_des = imagesy($im_des);

    # Création d'un filet noir autour de l'image
    imagerectangle($im_des, 0, 0, $w_des-1, $h_des-1, $color_noir);
    # Chargement de l'image du copyright
    $cpy = imagecreatefrompng(dirname(__FILE__).'/copyright');
    $w_cpy = imagesx($cpy);
    $h_cpy = imagesy($cpy);
    # Copie du copyright sur l'image finale en bas à droite
    imagecopy($im_des,$cpy,$w_des-$w_cpy-1, $h_des-$h_cpy-1,0,0,$w_cpy,$h_cpy);
    # On retourne l'image
    return $im_des;

  }

  # retourne le code html affichant une image et enregistrant un champ input invisible
  # contenant la clé publique
  public function getImgFormHtml($url_pab_web=''){
    $string = '';
    # On récupère une clée publique
    $public_key = $this->getPublicKey();
    # On génère quelque chose d'unique permettant de ne pas enreistrer les images en cache...
    $unique_string = mktime().md5(mktime().mt_rand(1,999)).mt_rand(1,999);
    # Formatage du code html avec les bonnes valeurs
    $string.="\n";
    $string.='<!-- START CODE OF PAB -->'."\n";
    $string.='<a href="http://www.natsimhan.com/PAB/" alt="PAB par Natsimhan : un anti-bot libre en PHP">'."\n";
    $string.='<img border="0" src="'.$url_pab_web.'pab.php?PublicKey='.$public_key.'&ImagePAB='.$unique_string.'" alt="Clée privée de PAB" />'."\n";
    $string.='</a>'."\n";
    $string.='<input type="hidden" name="PublicKey" value="'.$public_key.'" />'."\n";
    $string.='<!-- END CODE OF PAB -->'."\n";
    # On retourne le résultat pour qu'il puisse etre affiché
    return $string;
  }
  
  # Retourne TRUE si le couple de clé est bon, sinon retourne FALSE
  public function testKeys(){
    // --- On récupère la clée privée fournie par le user
    if(!isset($_POST['PrivateKey'])) return false;
    $private_key = $_POST['PrivateKey'];
    if(!settype($private_key,"string")) return false;
    // --- On récupère la clée publique définit dans le formulaire
    if(!isset($_POST['PublicKey'])) return false;
    $public_key = $_POST['PublicKey'];
    if(!settype($public_key,"string")) return false;
    # TRUE si le couple est bon, FALSE sinon
    if($this->_SenssibleCasse){
      return $this->getPrivateKey($public_key) === $private_key;
    }else{
      return strtolower($this->getPrivateKey($public_key)) === strtolower($private_key);
    }
  }
  /**
   * return width and height, offset [left, top] of a ttf character
   * @param int $size : the font size
   * @param int $angle : the angle
   * @param string $font : the font file
   * @param string $text : the character
   * @access private
   * @return array  of the width and height, left and top.
   **/
  private function myImagettfbbox($size, $angle, $font, $text){
    // Get the boundingbox from imagettfbbox(), which is correct when angle is 0
    $bbox = imagettfbbox($size, 0, $font, $text);

    // Rotate the boundingbox
    $angle = $angle/ 180 * pi();
    for ($i=0; $i<4; $i++){
      $x = $bbox[$i * 2];
      $y = $bbox[$i * 2 + 1];
      $bbox[$i * 2] = cos($angle) * $x - sin($angle) * $y;  // X
      $bbox[$i * 2 + 1] = sin($angle) * $x + cos($angle) * $y; // Y
    }
    // Variables which tells the correct width and height
    $bbox["left"] = 0- min($bbox[0],$bbox[2],$bbox[4],$bbox[6]);
    $bbox["top"] = 0- min($bbox[1],$bbox[3],$bbox[5],$bbox[7]);
    $bbox["width"] = max($bbox[0],$bbox[2],$bbox[4],$bbox[6]) -  min($bbox[0],$bbox[2],$bbox[4],$bbox[6]);
    $bbox["height"] = max($bbox[1],$bbox[3],$bbox[5],$bbox[7]) - min($bbox[1],$bbox[3],$bbox[5],$bbox[7]);

    return $bbox;
  }
  # Choisit une clée public aléatoirement et la retourne
  private function getPublicKey() {
    $public_key = "";
    # On remet le pointeur au début
    $tbk = $this->getTabPK();
    reset($tbk);
  	# On choisit une position aléatoire
    $indice = mt_rand(0, count($tbk)-1);
    # On parcours le tableau jusqu'à l'indice
    for($i=0;$i<$indice;$i++) next($tbk);
    # On récupère alors la clée se trouvant à cet indice
  	$public_key = key($tbk);
  	# On retourne la clée publique
    return $public_key;
	}

  # Retourne la clée privée correspondant à la clée publique donnée
  private function getPrivateKey($public_key) {
    if(!settype($public_key,"string")) return null;
  	return $this->getTabPK($public_key);
	}

  # Importation de la liste des associations des clées publiques/privées
  # Elles sont lues dans un fichier puis traduite en tableau dans la classe
  private function keyFileToTab(){
    # On déclare le tableau
    $pk = array();
    # Ouverture du fichier en lecture seule
    $fd = fopen ('./pk/.pk', "r");
    # Leture du fichier ligne après ligne
    while (!feof ($fd)) {
      $buffer = fgets($fd, 4096);
      $tab_buffer = preg_split('/\|/', $buffer, -1, PREG_SPLIT_NO_EMPTY);
      if(count($tab_buffer)>0){
        $pk[$tab_buffer[0]] = $tab_buffer[1];
      }
    }
    # Fermeture du fichier
    fclose ($fd);
    # Enregistrement du tableau dans la classe et de son nombre de valeur
    $this->setTabPK($pk);
  }

  protected function traceGrille($image){
    $color = imagecolorallocate($image, 0, 0, 0);

    $w = imagesx($image);
    $h = imagesy($image);

    # on fait d'abord les verticales
    $offset = mt_rand(10,30);
    $i_offset=$offset;
    $offset = mt_rand(0,1)==0 ? $offset : -$offset ;
    for($i=0;$i<$w;$i+=$i_offset){
      imageline ($image,$i,0,$i+$offset,$h,$color);
    }
    # puis les horizontales
    $offset = mt_rand(10,30);
    $i_offset=$offset;
    $offset = mt_rand(0,1)==0 ? $offset : -$offset ;
    for($i=0;$i<$w;$i+=$i_offset){
      imageline ($image,0,$i+$offset,$w,$i,$color);
    }
  }

  protected function setTabPK(array $value){
    $this->_TabPK = $value;
  }
  protected function getTabPK($index=null){
    if($index===null){
      return $this->_TabPK;
    }else{
      return $this->_TabPK[$index];
    }
  }
  public function setBordure($value){
    $this->_Bordure = (int)$value>0?$value:-$value;
  }
  public function getBordure(){
    return $this->_Bordure;
  }
  public function setMargeLettre($value){
    $this->_MargeLettre = (int)$value>0?$value:-$value;
  }
  public function getMargeLettre(){
    return $this->_MargeLettre;
  }
  public function setAngleMax($value){
    $this->_AngleMax = (int)$value;
  }
  public function getAngleMax(){
    return $this->_AngleMax;
  }
  public function setCopyResampled($value){
    $this->_CopyResampled = (bool)$value;
  }
  public function getCopyResampled(){
    return $this->_CopyResampled;
  }
  public function setFontType($value){
    $this->_FontType = (int)$value;
  }
  public function getFontType(){
    return $this->_FontType;
  }
  public function setSenssibleCasse($value){
    $this->_SenssibleCasse = (bool)$value;
  }
  public function getSenssibleCasse(){
    return $this->_SenssibleCasse;
  }
}
?>
